"""add inviter field for vfolder_invitations

Revision ID: 1fa6a31ea8e3
Revises: 26d0c387e764
Create Date: 2018-07-05 00:09:35.230704

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1fa6a31ea8e3"
down_revision = "26d0c387e764"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("vfolder_invitations", sa.Column("inviter", sa.String(length=256)))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("vfolder_invitations", "inviter")
    # ### end Alembic commands ###



"""add user model

Revision ID: 819c2b3830a9
Revises: 8e660aa31fe3
Create Date: 2019-05-02 00:21:43.704843

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from ai.backend.manager.models import UserRole
from ai.backend.manager.models.base import (
    GUID,
    EnumValueType,
    ForeignKeyIDColumn,
    IDColumn,
    convention,
)
from ai.backend.manager.models.user import PasswordColumn

# from ai.backend.manager.models import keypairs, users, UserRole


# revision identifiers, used by Alembic.
revision = "819c2b3830a9"
down_revision = "8e660aa31fe3"
branch_labels = None
depends_on = None


userrole_choices = list(map(lambda v: v.value, UserRole))
userrole = postgresql.ENUM(*userrole_choices, name="userrole")


def upgrade():
    metadata = sa.MetaData(naming_convention=convention)
    # partial table to be preserved and referred
    keypairs = sa.Table(
        "keypairs",
        metadata,
        sa.Column("user_id", sa.String(length=256), index=True),
        sa.Column("access_key", sa.String(length=20), primary_key=True),
        sa.Column("secret_key", sa.String(length=40)),
        sa.Column("is_active", sa.Boolean, index=True),
        sa.Column("is_admin", sa.Boolean, index=True),
        ForeignKeyIDColumn("user", "users.uuid", nullable=False),
    )
    # partial table to insert the migrated data
    users = sa.Table(
        "users",
        metadata,
        IDColumn("uuid"),
        sa.Column("username", sa.String(length=64), unique=True),
        sa.Column("email", sa.String(length=64), index=True, nullable=False, unique=True),
        sa.Column("password", PasswordColumn()),
        sa.Column("need_password_change", sa.Boolean),
        sa.Column("is_active", sa.Boolean, default=True),
        sa.Column("role", EnumValueType(UserRole), default=UserRole.USER),
    )

    userrole.create(op.get_bind())
    op.create_table(
        "users",
        sa.Column("uuid", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("username", sa.String(length=64), nullable=True),
        sa.Column("email", sa.String(length=64), nullable=False),
        sa.Column("password", PasswordColumn(), nullable=True),
        sa.Column("need_password_change", sa.Boolean(), nullable=True),
        sa.Column("first_name", sa.String(length=32), nullable=True),
        sa.Column("last_name", sa.String(length=32), nullable=True),
        sa.Column("description", sa.String(length=500), nullable=True),
        sa.Column("is_active", sa.Boolean(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column(
            "role",
            postgresql.ENUM(*userrole_choices, name="userrole", create_type=False),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_users")),
        sa.UniqueConstraint("username", name=op.f("uq_users_username")),
    )
    op.create_index(op.f("ix_users_email"), "users", ["email"], unique=True)
    op.add_column("keypairs", sa.Column("user", GUID(), nullable=True))
    op.create_foreign_key(op.f("fk_keypairs_user_users"), "keypairs", "users", ["user"], ["uuid"])

    # ### Create users based on keypair.user_id & associate keypairs.user to user record ###
    # Get all keypairs
    connection = op.get_bind()
    query = sa.select([
        keypairs.c.user_id,
        keypairs.c.access_key,
        keypairs.c.secret_key,
        keypairs.c.is_admin,
    ]).select_from(keypairs)
    results = connection.execute(query).fetchall()
    for keypair in results:
        email = keypair["user_id"]
        access_key = keypair["access_key"]
        is_admin = keypair["is_admin"]
        if email in [None, ""]:
            continue
        # Try to get a user whose email matches with current keypair's email
        query = (
            sa.select([users.c.uuid, users.c.role]).select_from(users).where(users.c.email == email)
        )
        user = connection.execute(query).first()
        if user:
            # Update user's role if current keypair is admin keypair
            user_uuid = user["uuid"]
            role = UserRole.ADMIN if is_admin else UserRole.USER
            if role == UserRole.ADMIN and user["role"] != UserRole.ADMIN:
                query = sa.update(users).values(role=UserRole.ADMIN).where(users.c.email == email)
                connection.execute(query)
        else:
            # Create new user (set username with email)
            role = UserRole.ADMIN if is_admin else UserRole.USER
            temp_password = keypair["secret_key"][:8]
            query = (
                sa.insert(users)
                .returning(users.c.uuid)
                .values(
                    username=email,
                    email=email,
                    password=temp_password,
                    need_password_change=True,
                    is_active=True,
                    role=role,
                )
            )
            user = connection.execute(query).first()
            user_uuid = user[0]
        # Update current keypair's `user` field with associated user's uuid.
        query = (
            sa.update(keypairs).values(user=user_uuid).where(keypairs.c.access_key == access_key)
        )
        connection.execute(query)

    # Make keypairs.user column NOT NULL.
    op.alter_column("keypairs", column_name="user", nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(op.f("fk_keypairs_user_users"), "keypairs", type_="foreignkey")
    op.drop_column("keypairs", "user")
    op.drop_index(op.f("ix_users_email"), table_name="users")
    op.drop_table("users")
    # ### end Alembic commands ###

    userrole.drop(op.get_bind())



from __future__ import annotations

import logging
from pprint import pformat, pprint

import click
import sqlalchemy as sa
from redis.asyncio.client import Pipeline, Redis
from tabulate import tabulate

from ai.backend.common import redis_helper
from ai.backend.common.docker import ImageRef
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.logging import BraceStyleAdapter

from ..models.image import ImageAliasRow, ImageRow
from ..models.image import rescan_images as rescan_images_func
from ..models.utils import connect_database
from .context import CLIContext, etcd_ctx, redis_ctx

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


async def list_images(cli_ctx, short, installed_only):
    # Connect to postgreSQL DB
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_readonly_session() as session,
        redis_ctx(cli_ctx) as redis_conn_set,
    ):
        displayed_items = []
        try:
            items = await ImageRow.list(session)
            # NOTE: installed/installed_agents fields are no longer provided in CLI,
            #       until we finish the epic refactoring of image metadata db.
            if installed_only:

                async def _build_scard_pipeline(redis: Redis) -> Pipeline:
                    pipe = redis.pipeline()
                    for item in items:
                        await pipe.scard(item.name)
                    return pipe

                installed_counts = await redis_helper.execute(
                    redis_conn_set.image, _build_scard_pipeline
                )
                installed_items = []

                async def _build_smembers_pipeline(redis: Redis) -> Pipeline:
                    pipe = redis.pipeline()
                    for item, installed_count in zip(items, installed_counts):
                        if installed_count > 0:
                            installed_items.append(item)
                            await pipe.smembers(item.name)
                    return pipe

                agents_per_installed_items = await redis_helper.execute(
                    redis_conn_set.image,
                    _build_smembers_pipeline,
                )
                for item, installed_agents in zip(installed_items, agents_per_installed_items):
                    formatted_installed_agents = " ".join(
                        map(lambda s: s.decode(), installed_agents)
                    )
                    if short:
                        displayed_items.append((
                            item.image_ref.canonical,
                            item.config_digest,
                            formatted_installed_agents,
                        ))
                    else:
                        print(f"{pformat(item)} @ {formatted_installed_agents}")
            else:
                for item in items:
                    if short:
                        displayed_items.append((item.image_ref.canonical, item.config_digest))
                    else:
                        pprint(item)
            if short:
                print(tabulate(displayed_items, tablefmt="plain"))
        except Exception:
            log.exception("An error occurred.")


async def inspect_image(cli_ctx, canonical_or_alias, architecture):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_readonly_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageRef(canonical_or_alias, ["*"], architecture),
                    canonical_or_alias,
                ],
            )
            pprint(await image_row.inspect())
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception:
            log.exception("An error occurred.")


async def forget_image(cli_ctx, canonical_or_alias, architecture):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageRef(canonical_or_alias, ["*"], architecture),
                    canonical_or_alias,
                ],
            )
            await session.delete(image_row)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception:
            log.exception("An error occurred.")


async def set_image_resource_limit(
    cli_ctx,
    canonical_or_alias,
    slot_type,
    range_value,
    architecture,
):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageRef(canonical_or_alias, ["*"], architecture),
                    canonical_or_alias,
                ],
            )
            await image_row.set_resource_limit(slot_type, range_value)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception:
            log.exception("An error occurred.")


async def rescan_images(cli_ctx: CLIContext, registry_or_image: str, local: bool) -> None:
    if not registry_or_image and not local:
        raise click.BadArgumentUsage("Please specify a valid registry or full image name.")
    async with (
        connect_database(cli_ctx.local_config) as db,
        etcd_ctx(cli_ctx) as etcd,
    ):
        try:
            await rescan_images_func(etcd, db, registry_or_image, local=local)
        except Exception:
            log.exception("An error occurred.")


async def alias(cli_ctx, alias, target, architecture):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        try:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageRef(target, ["*"], architecture),
                ],
            )
            await ImageAliasRow.create(session, alias, image_row)
        except UnknownImageReference:
            log.exception("Image not found.")
        except Exception:
            log.exception("An error occurred.")


async def dealias(cli_ctx, alias):
    async with (
        connect_database(cli_ctx.local_config) as db,
        db.begin_session() as session,
    ):
        alias_row = await session.scalar(
            sa.select(ImageAliasRow).where(ImageAliasRow.alias == alias),
        )
        if alias_row is None:
            log.exception("Alias not found.")
            return
        await session.delete(alias_row)
